import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.text.FieldPosition;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ResourceBundle;
import java.util.Locale;
import java.util.Date;
import java.util.TimeZone;
import java.util.Calendar;
import java.util.GregorianCalendar;

import java.awt.Dimension;
import java.awt.FontFormatException;
import java.awt.event.*;

import javax.swing.*;

import org.apache.log4j.Logger;

import com.taco.data.INoReturnObservableMap;
import com.taco.i18n.gui.FontUtilities;
import com.taco.i18n.gui.LocaleSelectorPanel;
import com.taco.text.GlobalUtilities;
import com.taco.util.ChainedResourceBundleFactory;
import com.taco.swinger.SwingInvokeProxyFactory;
import com.taco.swinger.text2gui.DispatchingComponentConverter;

/** A simple applet that converts time between two time zones. */
public class TimeZoneApplet extends JApplet implements PropertyChangeListener {
    public TimeZoneApplet() {
        super();                                            
    }
        
    /** An action run when the user hits the "Set Current Time" button.
     *  Normally, we would define such a class in the resource bundle, but
     *  because applets can't use BeanShell expressions, we have to define the
     *  class here.
     */    
    private class SetCurrentTimeAction extends AbstractAction {
        SetCurrentTimeAction() { super(); }
        
        public void actionPerformed(ActionEvent event) {
            TimeZoneApplet.this._setCurrentTime();                
        }        
    }

    /** An action run when the user hits the "Language ..." button. Normally, we
     *  would define such a class in the resource bundle, but because applets
     *  can't use BeanShell expressions, we have to define the class here.
     */
    private class SetLocaleAction extends AbstractAction {
        SetLocaleAction() { super(); }
        
        public void actionPerformed(ActionEvent event) {                                
            LocaleSelectorPanel localeSelectorPanel = 
              new LocaleSelectorPanel(_locale);              
            final boolean accepted[] = new boolean[1];  
            
            // We need to create panel here so we can get the locale from it
            // later.
            _argMap.putNoReturn("localeSelectorPanel", localeSelectorPanel,
              false);
            
            // Again, these actions would normally be defined in the bundle,
            // but since this is an applet, we need define them here.
            _argMap.putNoReturn("localeOKAction", new AbstractAction() {
                public void actionPerformed(ActionEvent event) {
                    // The dialog is named in the bundle, so we can a reference
                    // to it.
                    JDialog dialog = (JDialog) GlobalUtilities.getGlobal(
                      "localeDialog", _argMap);
                    accepted[0] = true;  
                    dialog.dispose();
                }
            }, false);
            _argMap.putNoReturn("localeCancelAction", new AbstractAction() {
                public void actionPerformed(ActionEvent event) {
                    // The dialog is named in the bundle, so we can a reference
                    // to it.                    
                    JDialog dialog = (JDialog) GlobalUtilities.getGlobal(
                      "localeDialog", _argMap);
                    accepted[0] = false;
                    dialog.dispose();
                }
            }, false);                    
                            
            try {                                                 
                JDialog dialog = (JDialog) 
                  DispatchingComponentConverter.DEFAULT_INSTANCE.
                  toComponent(_bundle, "localeDialog", _argMap);                                                     
                dialog.pack();
                
                FontUtilities.adjustFontsForLocale(dialog, _locale);
                                                
                dialog.show();
                
                if (accepted[0]) {
                    Locale selectedLocale = 
                      localeSelectorPanel.getSelectedLocale();
                    
                    if (selectedLocale != null) {
                        TimeZoneApplet.this.setLocale(selectedLocale);
                    }
                }
            } catch (Exception e) {
                _logger.warn("Can't display locale selector dialog", e);   
            }            
        }
    }    
    
    // Protect from entry until after init() is finished.
    protected boolean _inPropertyChange = true;
    
    /** Respond to keys in the argument map having their values changed. */ 
    public void propertyChange(PropertyChangeEvent evt) {
        //_logger.debug("Got property change: " + evt);
        
        if (_inPropertyChange) {
            return;
        }
        
        _inPropertyChange = true;
        
        try {
            boolean ispm = ((Boolean) _argMap.get("pm")).booleanValue();                                                                               
            int hour = ((Number) _argMap.get("hour")).intValue();
            int minute = ((Number) _argMap.get("minute")).intValue();
            int monthIndex = ((Number) _argMap.get("monthIndex")).intValue();
            int year = ((Number) _argMap.get("year")).intValue();
            int dayOfMonth = ((Number) _argMap.get("dayIndex")).intValue() + 1;
            
            int oldIndex = _sourceTimeZoneIndex; 
            _sourceTimeZoneIndex = 
             ((Integer) _argMap.get("fromTimeZoneIndex")).intValue(); 
            
            // Recover from GUI mistakes.
            if ((_sourceTimeZoneIndex < 0) || 
                (_sourceTimeZoneIndex >= _timeZones.length)) {
                _logger.warn("From time zone index invalid: " + 
                 _sourceTimeZoneIndex);
                _sourceTimeZoneIndex = oldIndex;                
            }
            
            TimeZone sourceTimeZone = _timeZones[_sourceTimeZoneIndex];
            
            oldIndex = _destTimeZoneIndex;
            _destTimeZoneIndex = 
             ((Integer) _argMap.get("toTimeZoneIndex")).intValue();

            // Recover from GUI mistakes.
            if ((_destTimeZoneIndex < 0) || 
                (_destTimeZoneIndex >= _timeZones.length)) {
                _logger.warn("To time zone index invalid: " + 
                 _destTimeZoneIndex);
                _destTimeZoneIndex = oldIndex;                
            }

            TimeZone destTimeZone = _timeZones[_destTimeZoneIndex];
                        
            Calendar source = new GregorianCalendar(sourceTimeZone);

            // Convert to military time.
            if (hour == 0) {
                // Military time, ignore am/pm
                ;
            } else if (hour == 12) {
                if (ispm) {
                    hour = 12;
                } else {
                    hour = 0;
                }
            } else if (hour <= 12) {
                if (ispm) {
                    hour += 12;
                }                
            } else {
                // Military time, ignore am/pm    
                ;
            }            
               
            source.set(year, monthIndex, dayOfMonth, hour, minute, 0);                        
            
            // Get UTC time.                                             
            Date date = source.getTime();
            
            DateFormat dateFormat = DateFormat.getDateTimeInstance(
              DateFormat.LONG, DateFormat.LONG, _locale);
            dateFormat.setTimeZone(sourceTimeZone);
                                        
            DateFormat dayOfWeekFormat = new SimpleDateFormat("EEEE", _locale);
                                                             
            _argMap.putNoReturn("dayOfWeek",
              dayOfWeekFormat.format(date, new StringBuffer(),
             new FieldPosition(0)).toString());                                                                                                      
              
            dateFormat.setTimeZone(destTimeZone);
                                                            
            _argMap.putNoReturn("destTime", dateFormat.format(date, 
              new StringBuffer(), new FieldPosition(0)).toString());
              
            _argMap.putNoReturn("destDayOfWeek",
             dayOfWeekFormat.format(date, new StringBuffer(),
             new FieldPosition(0)).toString());                                                                        
        } catch (Exception e) {
            _logger.warn("Update time exception", e);   
        } finally {
            _inPropertyChange = false;
        }
    }
     
    public void setLocale(Locale locale) {
        if ((_locale == null) || _locale.equals(locale)) {
            return;
        }                
        
        _locale = locale;

        _load();
        
        _logger.debug("Adjusting fonts");
        
        try {
            FontUtilities.adjustFontsForLocale(this, locale);
        } catch (FontFormatException ffe) {
            _logger.warn("Can't adjust fonts for locale " + locale);   
        }

        super.setLocale(locale);
        
        // Force a redraw.
        Dimension size = getSize();
        resize(size.width + 1, size.height);
        resize(size);                    
    }     
        
    public void init() {
        // Try to use the system look and feel.
        try {
            UIManager.setLookAndFeel(
             UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
            ; // ignore   
        }
        
        _initTimeZones();
        
        _sourceTimeZoneIndex = _defTimeZoneIndex;
        _destTimeZoneIndex = _defTimeZoneIndex;

        _load();
    }

    /** Set the current time when the applet is started again. */
    public void start() {
        if (!_hasStartedBefore) {
            _setCurrentTime();
            _hasStartedBefore = true;
        }
    }        
            
    /** Initialize the time zone array. */            
    protected TimeZone[] _initTimeZones() {
        String[] idArray = TimeZone.getAvailableIDs();
        TimeZone[] timeZones = new TimeZone[idArray.length];
        String[] displayNames = new String[idArray.length];
                
        TimeZone defTimeZone = TimeZone.getDefault();
        String defDisplayName = defTimeZone.getDisplayName();                             
                        
        int numZonesSet = 0;
                
        for (int i = 0; i < idArray.length; i++) {
            TimeZone timeZone = TimeZone.getTimeZone(idArray[i]);
            String displayName = timeZone.getDisplayName();            
                                        
            // Use the default time zone over an equivalent one so that
            // we are sure to have the default time zone in the array.
            if ((_defTimeZoneIndex == -1) &&                            
                displayName.equals(defDisplayName)) {                                                                                
                timeZone = defTimeZone;
                displayName = defDisplayName;
                _defTimeZoneIndex = numZonesSet;                             
            }                            
                                                                                        
            boolean equivalent = false;
            // Make sure the time zone isn't equivalent to one already set.
            for (int j = numZonesSet - 1; (j >= 0) && !equivalent; j--) {                                
                equivalent |= displayName.equals(displayNames[j]);
            }
             
            if (!equivalent) {
                timeZones[numZonesSet] = timeZone;
                displayNames[numZonesSet] = displayName;
                numZonesSet++;                                
            }                                    
        }                        
        
        _timeZones = new TimeZone[numZonesSet];
        System.arraycopy(timeZones, 0, _timeZones, 0, numZonesSet);
        
        // help gc
        idArray = null;
        timeZones = null;
        
        return _timeZones;
    }            

    protected void _load() {
        _inPropertyChange = true;                           
        _initBundle();        
        _initArgMap();
        _configureGUI();
        _addArgMapListeners();
        _inPropertyChange = false;
    }
            
    protected void _initBundle() {
        _bundle = ChainedResourceBundleFactory.DEFAULT_INSTANCE.getBundle(
           TimeZoneApplet.class.getName() + "ResourceBundle", _locale,
           TimeZoneApplet.class.getClassLoader());   
    }
    
    /** Create the argument map and put constant values into it. */
    protected void _initArgMap() {
        _argMap = SwingInvokeProxyFactory.makeSwingInvokeProxyMap();
        _argMap.putNoReturn("setLocaleAction", new SetLocaleAction(), false);
        _argMap.putNoReturn("setCurrentTimeAction", 
          new SetCurrentTimeAction(), false);                                       
    }                
                    
    protected void _configureGUI() {                
        String[] timeZoneNames = new String[_timeZones.length];
        for (int i = 0; i < _timeZones.length; i++) {
            timeZoneNames[i] = _timeZones[i].getDisplayName(false,
            TimeZone.LONG, _locale);            
        }                
        _argMap.putNoReturn("timeZoneNames", timeZoneNames, false);
        
        _setCurrentTime();

        DispatchingComponentConverter.DEFAULT_INSTANCE.configureComposite(
          "japplet", this, "tzApplet", _bundle, _argMap);                                          
    }   

    /** Update the argument map so that its values reflect the current time. */
    protected void _setCurrentTime() {
        // Prevent this method from responding while we change the argument
        // map, since we are recalculating the date fields here.
        boolean oldPropertyChange = _inPropertyChange;
        _inPropertyChange = true;        
        
        // Set fields with current date/time.                
        Calendar calendar = new GregorianCalendar(
          _timeZones[_sourceTimeZoneIndex]);
        _argMap.putNoReturn("monthIndex", 
          new Integer(calendar.get(Calendar.MONTH)));
        _argMap.putNoReturn("dayIndex", 
          new Integer(calendar.get(Calendar.DATE) - 1));
        _argMap.putNoReturn("year", 
          new Integer(calendar.get(Calendar.YEAR)));
                                      
        int hour = calendar.get(Calendar.HOUR);
        if (hour == 0) {
            hour = 12;
        }
        _argMap.putNoReturn("hour",  new Integer(hour));
        _argMap.putNoReturn("minute", 
          new Integer(calendar.get(Calendar.MINUTE)));
        _argMap.putNoReturn("pm", 
          new Boolean(calendar.get(Calendar.AM_PM) == Calendar.PM));     
        _argMap.putNoReturn("fromTimeZoneIndex", 
          new Integer(_sourceTimeZoneIndex));        
        _argMap.putNoReturn("toTimeZoneIndex", 
          new Integer(_destTimeZoneIndex));

        Date date = calendar.getTime();
        DateFormat dayOfWeekFormat = new SimpleDateFormat("EEEE", 
          _locale);
        dayOfWeekFormat.setTimeZone(_timeZones[_sourceTimeZoneIndex]);  
        String dayOfWeek = dayOfWeekFormat.format(calendar.getTime(), 
         new StringBuffer(), new FieldPosition(0)).toString();   
        _argMap.putNoReturn("dayOfWeek", dayOfWeek);
        
        dayOfWeekFormat.setTimeZone(_timeZones[_destTimeZoneIndex]);                        
        _argMap.putNoReturn("destDayOfWeek", dayOfWeek);
                
        DateFormat dateFormat = DateFormat.getDateTimeInstance(
          DateFormat.LONG, DateFormat.LONG, _locale);
        dateFormat.setTimeZone(_timeZones[_destTimeZoneIndex]);                            
        _argMap.putNoReturn("destTime", dateFormat.format(date, 
          new StringBuffer(), new FieldPosition(0)).toString());

        _inPropertyChange = oldPropertyChange;                                                                                        
    }
    
    /** Add this instance as a listener for options that might be changed
     *  by the user.
     */
    protected void _addArgMapListeners() {
        _argMap.addPropertyChangeListener("fromTimeZoneIndex", this);
        _argMap.addPropertyChangeListener("toTimeZoneIndex", this);
        _argMap.addPropertyChangeListener("monthIndex", this);
        _argMap.addPropertyChangeListener("dayIndex", this);
        _argMap.addPropertyChangeListener("year", this);        
        _argMap.addPropertyChangeListener("hour", this);
        _argMap.addPropertyChangeListener("minute", this);
        _argMap.addPropertyChangeListener("pm", this);
    }                          
                         
    protected INoReturnObservableMap _argMap = null;
    protected ResourceBundle _bundle = null;    
    protected TimeZone[] _timeZones = null;    
    protected int _sourceTimeZoneIndex = -1;
    protected int _destTimeZoneIndex = -1;
    protected int _defTimeZoneIndex = -1;
    protected boolean _hasStartedBefore = false;
    protected Locale _locale = Locale.getDefault();
          
    private static final Logger _logger =
      Logger.getLogger(TimeZoneApplet.class);
}

